// --------------------------------------------------------------------
// This part of the code was generated by including the file "seq-proc.v"
// --------------------------------------------------------------------
// Verilog representation of SEQ
// This file connects together all of the blocks to create the processor

// The tag INSERT-LOGIC-HERE (but with underscores rather than dashes)
// is used to indicate where the logic derived from the HCL code should be inserted.

// The processor can run in 5 different modes:
// RUN: Normal operation
// RESET: sets PC to 0, clears all pipe registers, initializes condition codes
// DOWNLOAD: Download bytes from controller into memory
// UPLOAD: Upload bytes from memory to controller
// STATUS: Upload other status information to controller

// Processor module
module processor(mode, udaddr, idata, odata, status, clock);
  input  [2:0]  mode;   // Signal operating mode to processor
  input  [31:0] udaddr; // Upload/download address
  input  [31:0] idata;  // Download data word
  output [31:0] odata;  // Upload data word
  output [2:0]  status; // Exception status
  input clock;          // Clock input

// Define modes
  parameter RUN_MODE = 0;      // Normal operation
  parameter RESET_MODE = 1;    // Resetting processor;
  parameter DOWNLOAD_MODE = 2; // Transfering to memory
  parameter UPLOAD_MODE = 3;   // Reading from memory
  parameter STATUS_MODE = 4;   // Uploading register & other status information

// Constant values

  // Instruction functions
  parameter INOP =    4'h0;
  parameter IHALT =   4'h1;
  parameter IRRMOVL = 4'h2;
  parameter IIRMOVL = 4'h3;
  parameter IRMMOVL = 4'h4;
  parameter IMRMOVL = 4'h5;
  parameter IOPL =    4'h6;
  parameter IJXX =    4'h7;
  parameter ICALL =   4'h8;
  parameter IRET =    4'h9;
  parameter IPUSHL =  4'hA;
  parameter IPOPL =   4'hB;
  parameter IIADDL =  4'hC;
  parameter ILEAVE =  4'hD;

  // Register IDs
  parameter RESP =  4'h4;
  parameter REBP =  4'h5;
  parameter RNONE = 4'h8;

  // ALU operations
  parameter ALUADD = 4'h0;

  // Exception conditions
  parameter EBUB = 3'h0; // Never happens
  parameter EAOK = 3'h1;
  parameter EHLT = 3'h2;
  parameter EADR = 3'h3;
  parameter EINS = 3'h4;
  parameter EPIP = 3'h5; // Never happens

// Control signals
  wire resetting = (mode == RESET_MODE);
  wire uploading = (mode == UPLOAD_MODE);
  wire downloading = (mode == DOWNLOAD_MODE);
  wire running = (mode == RUN_MODE);
  wire getting_status = (mode == STATUS_MODE);

// Fetch stage
  wire f_ok;
  wire [31:0] pc;
  wire [47:0] instr;
  wire [3:0] icode;
  wire [3:0] ifun;
  wire [3:0] rA;
  wire [3:0] rB;
  wire [31:0] valC;
  wire [31:0] valP;
  wire need_regids;
  wire need_valC;
  wire instr_valid;
// Decode stage
  wire [31:0] valA;
  wire [31:0] valB;
  wire [3:0] dstE;
  wire [3:0] dstM;
  wire [3:0] srcA;
  wire [3:0] srcB;
// Execute stage
  wire [31:0] aluA;
  wire [31:0] aluB;
  wire set_cc;
  wire [2:0] cc;
  wire [2:0] new_cc;
  wire [3:0] alufun;
  wire Bch;
  wire [31:0] valE;
// Memory stage
  wire m_ok;
  wire [31:0] mem_addr;
  wire [31:0] mem_data;
  wire mem_read;
  wire mem_write;
  wire [31:0] valM;
// PC selection
  wire [31:0] new_pc;
// Exception status
  wire [2:0] exc;

// Debugging logic
  wire [31:0] eax, ecx, edx, ebx, esp, ebp, esi, edi;
  wire zf = cc[2];
  wire sf = cc[1];
  wire of = cc[0];

// Control signals
  assign status = exc;
  assign odata = 
   // When getting status, get either register or special status value
    getting_status ?  
     (udaddr == 0 ? eax :
      udaddr == 4 ? ecx :
      udaddr == 8 ? edx :
      udaddr == 12 ? ebx :
      udaddr == 16 ? esp :
      udaddr == 20 ? ebp :
      udaddr == 24 ? esi :
      udaddr == 28 ? edi :
      udaddr == 32 ? cc :
      udaddr == 36 ? pc : 0)
    : valM;

// The blocks 

// Fetch stage
  cenrreg #(32) p(pc, new_pc, running & exc == EAOK, resetting, 0, clock);  // Program counter
  split splt(instr[7:0], icode, ifun);
  align aln(instr[47:8], need_regids, rA, rB, valC);
  pc_increment pci(pc, need_regids, need_valC, valP);

// Decode stage
  regfile regf(dstE, valE, dstM, valM, srcA, valA, srcB, valB, 
     resetting, clock,
     eax, ecx, edx, ebx, esp, ebp, esi, edi);

// Execute stage
  alu alu(aluA, aluB, alufun, valE, new_cc);
  cc  ccreg(cc, new_cc, running & set_cc & exc == EAOK,
            reset, clock);
  bcond bc(icode, ifun, cc, Bch);

// Memory stage
  bmemory m(running & mem_write | downloading,
           downloading ? udaddr : mem_addr,
           downloading ? idata : mem_data,
           running & mem_read | uploading,
           uploading ? udaddr : mem_addr,
           valM, m_ok, pc, instr, f_ok, clock);

// Operational control
  assign exc =
    ~f_ok | ~m_ok ?  EADR :
    ~instr_valid  ?  EINS :
    icode == IHALT ? EHLT :
    EAOK;

// Glue logic
// --------------------------------------------------------------------
// The following code is generated from the HCL description of the
// processor control using a program hcl2v
// --------------------------------------------------------------------
// INSERT_LOGIC_HERE
// --------------------------------------------------------------------
// End of code generated by hcl2v
// --------------------------------------------------------------------
endmodule
// --------------------------------------------------------------------
// End of file "seq-proc.v"
// --------------------------------------------------------------------
