/* Implementation of floating point functions using unsigned operations */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "float-functions.h"

/*************** Helper functions ******************************************/

float_bits get_bits(float f) {
    union {
	float fv;
	float_bits bv;
    } u;
    u.fv = f;
    return u.bv;
}

float get_float(float_bits b) {
    union {
	float fv;
	float_bits bv;
    } u;
    u.bv = b;
    return u.fv;
}

int isNaN(float_bits f) {
    unsigned exp =  f>>23 & 0xFF;
    unsigned frac = f & 0x7FFFFF;
    return exp == 0xFF && frac != 0;
}

void show_bits(float_bits f)
{
    unsigned sign = f>>31;
    unsigned exp =  f>>23 & 0xFF;
    unsigned frac = f     & 0x7FFFFF;
    printf("Bits: 0x%.8x, (sign %d, exp 0x%.2x, frac 0x%.6x).  Value %.8f\n",
	   f, sign, exp, frac, get_float(f));
}

/************************* Functions and Test Functions *********************/


/* $begin float_denorm_zero */
/* If f is denorm, return 0.  Otherwise, return f */
float_bits float_denorm_zero(float_bits f) {
    /* Decompose bit representation into parts */
    unsigned sign = f>>31;
    unsigned exp =  f>>23 & 0xFF;
    unsigned frac = f     & 0x7FFFFF;
    if (exp == 0) {
	/* Denormalized.  Set fraction to 0 */
        frac = 0;
    }
    /* Reassemble bits */
    return (sign << 31) | (exp << 23) | frac;
}
/* $end float_denorm_zero */

/* $begin float_twice_solve */
/* Compute 2*f.  If f is NaN, then return f. */
float_bits float_twice(float_bits f) {
    unsigned sign = f>>31;
    unsigned exp =  f>>23 & 0xFF;
    unsigned frac = f     & 0x7FFFFF;
    if (exp == 0) {
	/* Denormalized.  Must double fraction */
	frac = 2*frac;
	if (frac > 0x7FFFFF) {
	    /* Result normalized */
	    frac = frac & 0x7FFFFF; /* Chop off leading bit */
	    exp = 1;
	}
    } else if (exp < 0xFF) {
	/* Normalized.  Increase exponent */
	exp++;
	if (exp == 0xFF) {
	    /* Infinity */
	    frac = 0;
	}
    } else if (frac != 0) {
	/* NaN */
	return f;
    }
    /* Infinity does not require any changes */
    return (sign << 31) | (exp << 23) | frac;
}
/* $end float_twice_solve */

float_bits test_float_twice(float_bits f) {
    if (isNaN(f))
	return f;
    return get_bits(2*get_float(f));
}

/* $begin float_half_solve */
/* Compute 0.5*f.  If f is NaN, then return f. */
float_bits float_half(float_bits f) {
    unsigned sign = f>>31;
    unsigned exp =  f>>23 & 0xFF;
    unsigned frac = f     & 0x7FFFFF;
    /* Only roundup case will be when rounding to even */
    unsigned roundup = (frac & 0x3) == 3;
    if (exp == 0) {
	/* Denormalized.  Must halve fraction */
	frac = (frac >> 1) + roundup;
    } else if (exp < 0xFF) {
	/* Normalized.  Decrease exponent */
	exp--;
	if (exp == 0) {
	    /* Denormalize, add back leading one */
	    frac = (frac >> 1) + roundup + 0x400000;
	}
    } else if (frac != 0) {
	/* NaN */
	return f;
    }
    /* Infinity does not require any changes */
    return (sign << 31) | (exp << 23) | frac;
}
/* $end float_half_solve */

float_bits test_float_half(float_bits f) {
    if (isNaN(f))
	return f;
    return get_bits(0.5*get_float(f));
}

/* $begin float_absval_solve */
/* Compute |f|.  If f is NaN, then return f. */
float_bits float_absval(float_bits f) {
    unsigned exp =  f>>23 & 0xFF;
    unsigned frac = f     & 0x7FFFFF;
    unsigned mask = 1 << 31;
    unsigned absval = f & ~mask;
    if (exp == 0xFF && frac != 0)
	/* NaN */
	return f;
    return absval;
}
/* $end float_absval_solve */

float_bits test_float_absval(float_bits f) {
    float fval = get_float(f);
    if (isNaN(f))
	return f;
    if (fval == 0.0)
	return get_bits(0.0);
    if (fval > 0.0)
	return get_bits(fval);
    return get_bits(-fval);
}

/* $begin float_negate_solve */
/* Compute -f.  If f is NaN, then return f. */
float_bits float_negate(float_bits f) {
    unsigned exp =  f>>23 & 0xFF;
    unsigned frac = f     & 0x7FFFFF;
    unsigned mask = 1 << 31;
    unsigned neg = f ^ mask;
    if (exp == 0xFF && frac != 0)
	/* NaN */
	return f;
    return neg;
}
/* $end float_negate_solve */

float_bits test_float_negate(float_bits f) {
    float fval = get_float(f);
    if (isNaN(f))
	return f;
    return get_bits(-fval);
}

/* $begin float_i2f_solve */
/* Compute (float) i */
float_bits float_i2f(int i) {
    unsigned sign = (i < 0);
    unsigned ai = (i < 0) ? -i : i;
    unsigned exp = 127+31;
    unsigned residue;
    unsigned frac = 0;
    if (ai == 0) {
	exp = 0;
	frac = 0;
    } else {
	/* Normalize so that msb = 1 */
	while ((ai & (1<<31)) == 0) {
	    ai = ai << 1;
	    exp--;
	}
	/* Now have Bit 31 = MSB (becomes implied leading one)
	   Bits 8-30 are tentative fraction,
	   Bits 0-7 require rounding.
	*/
	residue = ai & 0xFF;
	frac = (ai >> 8) & 0x7FFFFF; /* 23 bits */
	if (residue > 0x80 || (residue == 0x80 && (frac & 0x1))) {
	    /* Round up */
	    frac ++;
	    /* Might need to renormalize */
	    if (frac > 0x7FFFFF) {
		frac = (frac & 0x7FFFFF) >> 1;
		exp++;
	    }
	}
    }
    return (sign << 31) | (exp << 23) | frac;
}
/* $end float_i2f_solve */

float_bits test_float_i2f(int i) {
    return get_bits((float) i);
}

/* $begin float_f2i_solve */
/*
 * Compute (int) f.
 * If conversion causes overflow or f is NaN, return 0x80000000
 */
int float_f2i(float_bits f) {
    unsigned sign = f >> 31;
    unsigned exp = (f >> 23) & 0xFF;
    unsigned frac = f & 0x7FFFFF;
    /* Create normalized value with leading one inserted,
       and rest of significand in bits 8--30.
    */
    unsigned val = 0x80000000u + (frac << 8);
    if (exp < 127) {
	/* Absolute value is < 1 */
	return (int) 0;
    }
    if (exp > 158)
	/* Overflow */
	return (int) 0x80000000u;
    /* Shift val right */
    val = val >> (158 - exp);
    /* Check if out of range */
    if (sign) {
	/* Negative */
	return val > 0x80000000u ? (int) 0x80000000u : -(int) val;
    } else {
	/* Positive */
	return val > 0x7FFFFFFF ? (int) 0x80000000u : (int) val;
    }
}
/* $end float_f2i_solve */

int test_float_f2i(float_bits f) {
    return (int) get_float(f);
}

/**************************************************************************************/

/* Testing code */
typedef unsigned (*fun_t)(unsigned);
typedef struct {
    char *name;
    fun_t fun;
    fun_t test_fun;
    int int_arg;
    int int_result;
} entry_struct;

entry_struct table[] = {
    {"twice", float_twice, test_float_twice, 0, 0},
    {"half", float_half, test_float_half, 0, 0},
    {"absval", float_absval, test_float_absval, 0, 0},
    {"negate", float_negate, test_float_negate, 0, 0},
    {"i2f", (fun_t) float_i2f, (fun_t) test_float_i2f, 1, 0},
    {"f2i", (fun_t) float_f2i, (fun_t) test_float_f2i, 0, 1},
    {NULL, NULL, NULL}
};

int run_test(char *name, fun_t fun, fun_t test_fun, int int_arg, int int_result, int limit) {
    unsigned u = 0;
    printf("Function %s ", name);
    fflush(stdout);
    do {
	unsigned fval = fun(u);
	unsigned tval = test_fun(u);
	if (fval != tval) {
	    limit--;
	    printf("... mismatch:\n");
	    printf("  Argument: ");
	    if (int_arg)
		printf("%d (0x%.8x)\n", (int) u, u);
	    else
		show_bits(u);
	    printf("  Test result: ");
	    if (int_result)
		printf("%d (0x%.8x)\n", (int) tval, tval);
	    else
		show_bits(tval);
	    printf("  Function result: ");
	    if (int_result)
		printf("%d (0x%.8x)\n", (int) fval, fval);
	    else
		show_bits(fval);
	}
	u++;
    } while (u != 0 && limit > 0);
    if (limit > 0)
	printf("... OK\n");
    return limit > 0;
}

int main(int argc, char *argv[]) {
    int i;
    int limit = 10;
    for (i = 1; i < argc; i++) {
	char *name = argv[i];
	int index;
	for (index = 0; table[index].name; index++)
	    if (strcmp(table[index].name, name) == 0) {
		run_test(name, table[index].fun, table[index].test_fun,
			 table[index].int_arg, table[index].int_result, limit);
		break;
	    }
    }
    return 0;
}
